#!/usr/bin/env python3

# REQUIRES: biopython, numpy

# stdlib imports
import os, sys
import math
from math import pi, atan2, sin, cos

# library dependencies

try:
    import numpy as np
    import numpy.linalg as npla
except:
    print("[error] Failed to load NumPy! Please install NumPy (https://numpy.org/install/).")
    exit(1)

try:
    from Bio.PDB import PDBParser
    from Bio.PDB import Superimposer
    from Bio.PDB.internal_coords import *
except:
    print("[error] Failed to load BioPython! Please install BioPython (https://biopython.org/wiki/Download).")
    exit(1)


parser = PDBParser()
superimposer = Superimposer()

reference_path = sys.argv[1]
mut_paths = list(sys.argv[2:])

reference = parser.get_structure(reference_path, reference_path)
muts = {x:parser.get_structure(x, x) for x in mut_paths}

reference.atom_to_internal_coordinates()

print(f'reference,variant,ca_rmsd,mean_abs_diff_chi1,num_different_chi1,residues_with_different_chi1')
for k,v in muts.items():
    mutofname = k+".csv"


    # calculate global RMSD
    superimposer.set_atoms(
            [x for x in reference.get_atoms() if x.get_id() == "CA"],
            [x for x in         v.get_atoms() if x.get_id() == "CA"],
            )
    rmsd = superimposer.rms
    
    superimposer.apply(list(v.get_atoms()))

    v.atom_to_internal_coordinates()

    ref_res = [x for x in reference.get_residues()]
    ref_resn = [x.get_resname() for x in reference.get_residues()]
    v_res = [x for x in v.get_residues()]
    v_resn = [x.get_resname() for x in v.get_residues()]

    # collect chi1 torsion angles (degrees)
    ref_chi = [x.internal_coord.get_angle('chi1') or 0. for x in reference.get_residues()]
    v_chi = [x.internal_coord.get_angle('chi1') or 0. for x in v.get_residues()]
    
    goodidx = [(ref_resn[i]==v_resn[i]) and (ref_chi[i] != v_chi[i]) for i in range(len(ref_chi))]

    xref_chi = [ref_chi[i] for i in range(len(ref_chi)) if goodidx[i]]
    xv_chi   = [v_chi[i]   for i in range(len(v_chi  )) if goodidx[i]]

    # angles need to be converted to radians for calculation, then back to degrees for output
    
    allchidiff = np.array(ref_chi)-np.array(v_chi)
    allchidiff = np.array([( abs(atan2(sin(x*pi/180.), cos(x*pi/180.))) ) for x in allchidiff])

    chidiff = np.array(xref_chi)-np.array(xv_chi)
    chidiff = np.array([( abs(atan2(sin(x*pi/180.), cos(x*pi/180.))) ) for x in chidiff])
    chidiffmean = np.mean(chidiff)
    chidiffmean = chidiffmean * 180. / pi
    num_different_chi1 = goodidx.count(True)
    
    residues_with_different_chi1 = [list(v.get_residues())[i] for i in range(len(ref_chi)) if goodidx[i]]
    residues_with_different_chi1 = [x.get_resname().title()+str(x.get_id()[1]) for x in residues_with_different_chi1]
    str_residues_with_different_chi1 = "+".join(residues_with_different_chi1)

    print(f'{reference_path},{k},{rmsd:.2f},{chidiffmean:.1f},{num_different_chi1},{str_residues_with_different_chi1}')

    with open(mutofname, 'w') as f:
        print("reference,variant,residue_number,residue_name,ca_position_difference,chi1,chi1_absolute_difference,was_sidechain_moved,was_residue_mutated", file=f)
        for i in range(len(ref_chi)):
            refr = ref_res[i]
            vr = v_res[i]
            if ("CA" in refr) and ("CA" in vr):
                caposdiff = npla.norm(refr["CA"].get_vector() - vr["CA"].get_vector())
            else:
                caposdiff = 0.

            cur_v_chi = v_chi[i]
            cur_chi_diff = allchidiff[i]*180./pi
            ismoved = goodidx[i]
            ismut = (ref_resn[i]!=v_resn[i])

            print(f"{reference_path},{k},{vr.get_id()[1]},{vr.get_resname().title()},{caposdiff},{cur_v_chi},{cur_chi_diff},{ismoved},{ismut}", file=f)
